# -*- coding: utf-8 -*-
from ..parsers import (
    BaseParser,
    OpenAIParser,
    LLaMACPPParser,
    MistralAILargeParser,
)


class BaseAssistant(BaseParser):
    def __init__(self, target_cls, *args, **kwargs):
        super().__init__(target_cls, *args, **kwargs)
        self.system_prompt = getattr(self.target_cls, "system_prompt", "")
        self.prompt = getattr(self.target_cls, "prompt", "")

    def process(self, **kwargs):
        system_prompt = self.system_prompt.format(**kwargs)
        prompt = self.prompt.format(**kwargs)

        self.model_wrapper.system_prompt = system_prompt

        completion = self.model_wrapper.ask(
            prompt, schema=self.target_json_schema, **self.completion_kwargs
        )
        instance = self.deserialize(completion.choices[0].message.content)
        return instance


class OpenAIAssistant(BaseAssistant, OpenAIParser):
    def __init__(self, target_cls, model="gpt-3.5-turbo", *args, **kwargs):
        super().__init__(target_cls, model=model, *args, **kwargs)


class LLaMACPPAssistant(BaseAssistant, LLaMACPPParser):
    def __init__(
        self,
        target_cls,
        model="mistral",
        llama_cpp_kwargs=None,
        *args,
        **kwargs
    ):
        super().__init__(
            target_cls,
            model=model,
            llama_cpp_kwargs=llama_cpp_kwargs,
            *args,
            **kwargs
        )


class MistralAILargeAssistant(BaseAssistant, MistralAILargeParser):
    def __init__(
        self, target_cls, model="mistral-large-latest", *args, **kwargs
    ):
        super().__init__(target_cls, model=model, *args, **kwargs)



from .base import OpenAIAssistant, LLaMACPPAssistant, MistralAILargeAssistant
from .summarizers import Summarizer, SimpleSummary, DenserSummaryCollection
from .verifiers import (
    QuestionCollection,
    Doubter,
    AnswerConsistency,
    ConsistencyVerifier,
)
from .analysts import Answer, Analyst


__all__ = [
    "OpenAIAssistant",
    "LLaMACPPAssistant",
    "MistralAILargeAssistant",
    "Summarizer",
    "SimpleSummary",
    "DenserSummaryCollection",
    "QuestionCollection",
    "Doubter",
    "AnswerConsistency",
    "ConsistencyVerifier",
    "Answer",
    "Analyst",
]



# -*- coding: utf-8 -*-
from mistralai.client import MistralClient
from mistralai.models.chat_completion import ChatMessage

from dataclasses import dataclass

from .base import (
    LLMBase,
    ChatCompletion,
)

from ..settings import (
    MISTRAL_API_KEY,
    USE_AZURE_AI_MISTRAL_LARGE,
    AZURE_AI_MISTRAL_LARGE_ENDPOINT,
    AZURE_AI_MISTRAL_LARGE_KEY,
)


def create_chat_completion(
    model, messages, temperature, max_tokens=1000, **kwargs
):
    extra_args = {}
    if USE_AZURE_AI_MISTRAL_LARGE:
        endpoint = AZURE_AI_MISTRAL_LARGE_ENDPOINT
        api_key = AZURE_AI_MISTRAL_LARGE_KEY
        model = "azureai"
        extra_args["endpoint"] = endpoint
    else:
        api_key = MISTRAL_API_KEY
        model = "mistral-large-latest"

    client = MistralClient(api_key=api_key, **extra_args)

    completion = client.chat(model=model, messages=messages, **kwargs)

    return completion


@dataclass
class MistralAILarge(LLMBase):
    name: str = "mistral-large-latest"
    system_prompt: str = "You are a helpful assistant"

    def __enter__(self):
        # No special initialization required as we are using API
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # No special cleanup required as we are using API
        pass

    @property
    def ctx_size(self):
        return 32_000

    def ask(
        self,
        prompt,
        history=None,
        schema=None,
        temperature=0,
    ):
        max_tokens = self.sanitize_prompt(
            prompt=prompt,
            history=history,
            schema=schema,
        )

        messages = [
            ChatMessage(
                **{
                    "role": "system",
                    "content": self.system_prompt,
                }
            ),
        ]
        if history:
            messages += [ChatMessage(**msg) for msg in history]

        messages.append(
            ChatMessage(
                **{
                    "role": "user",
                    "content": prompt,
                }
            ),
        )

        kwargs = {}
        if schema:
            tools = [
                {
                    "type": "function",
                    "function": {
                        "name": "PublishAnswer",
                        "description": "Publish the answer",
                        "parameters": schema,
                    },
                }
            ]
            tool_choice = "any"
            kwargs = {
                "tools": tools,
                "tool_choice": tool_choice,
            }

        completion = create_chat_completion(
            model=self.name,
            messages=messages,
            temperature=temperature,
            max_tokens=max_tokens,
            **kwargs,
        )

        return ChatCompletion.parse(completion.dict())



# -*- coding: utf-8 -*-
import re
import json
import dataclasses
import typing
from enum import Enum

from llama_cpp.llama_grammar import LlamaGrammar


def to_json_schema(datacls):
    def get_type(field_type):
        if field_type == int:
            return {"type": "integer"}
        elif field_type == str:
            return {"type": "string"}
        elif field_type == bool:
            return {"type": "boolean"}
        elif field_type == float:
            return {"type": "number"}
        elif field_type == complex:
            return {"type": "string", "format": "complex-number"}
        elif field_type == bytes:
            return {"type": "string", "contentEncoding": "base64"}
        elif field_type == tuple:
            return {"type": "array", "items": {}}
        elif field_type.__name__ == "set":
            return {
                "type": "array",
                "uniqueItems": True,
                "items": get_type(field_type.__args__[0]),
            }
        elif field_type.__name__ == "list":
            return {
                "type": "array",
                "items": get_type(field_type.__args__[0]),
            }
        elif isinstance(field_type, typing._GenericAlias):
            if field_type._name == "List":
                return {
                    "type": "array",
                    "items": get_type(field_type.__args__[0]),
                }
            elif field_type._name == "Dict":
                return {
                    "type": "object",
                    "additionalProperties": get_type(field_type.__args__[1]),
                }
        elif dataclasses.is_dataclass(field_type):
            return to_json_schema(field_type)
        elif issubclass(field_type, Enum):
            return {
                "type": "string",
                "enum": list(field_type.__members__.keys()),
            }
        else:
            return {"type": "object"}

    properties = {}
    required = []
    for field in dataclasses.fields(datacls):
        properties[field.name] = get_type(field.type)
        if (
            field.default == dataclasses.MISSING
            and field.default_factory == dataclasses.MISSING
        ):
            required.append(field.name)

    return {"type": "object", "properties": properties, "required": required}


def from_dict(cls, data):
    if dataclasses.is_dataclass(cls):
        field_types = {f.name: f.type for f in dataclasses.fields(cls)}
        return cls(
            **{k: from_dict(field_types[k], v) for k, v in data.items()}
        )
    elif cls.__name__ == "list":
        return [from_dict(cls.__args__[0], v) for v in data]
    elif cls.__name__ == "set":
        return set([from_dict(cls.__args__[0], v) for v in data])
    elif isinstance(cls, typing._GenericAlias):
        if cls._name == "List":
            return [from_dict(cls.__args__[0], v) for v in data]
        elif cls._name == "Dict":
            return {k: from_dict(cls.__args__[1], v) for k, v in data.items()}
    elif issubclass(cls, Enum):
        return getattr(cls, data)
    else:
        return data


SPACE_RULE = '" "?'

PRIMITIVE_RULES = {
    "boolean": '("true" | "false") space',
    "number": '("-"? ([0-9] | [1-9] [0-9]*)) ("." [0-9]+)? ([eE] [-+]? [0-9]+)? space',
    "integer": '("-"? ([0-9] | [1-9] [0-9]*)) space',
    "string": r""" "\"" (
        [^"\\] |
        "\\" (["\\/bfnrt] | "u" [0-9a-fA-F] [0-9a-fA-F] [0-9a-fA-F] [0-9a-fA-F])
      )* "\"" space """,
    "null": '"null" space',
}

INVALID_RULE_CHARS_RE = re.compile(r"[^a-zA-Z0-9-]+")
GRAMMAR_LITERAL_ESCAPE_RE = re.compile(r'[\r\n"]')
GRAMMAR_LITERAL_ESCAPES = {"\r": "\\r", "\n": "\\n", '"': '\\"'}


class SchemaConverter:
    def __init__(self, prop_order):
        self._prop_order = prop_order
        self._rules = {"space": SPACE_RULE}

    def _format_literal(self, literal):
        escaped = GRAMMAR_LITERAL_ESCAPE_RE.sub(
            lambda m: GRAMMAR_LITERAL_ESCAPES.get(m.group(0)),
            json.dumps(literal),
        )
        return f'"{escaped}"'

    def _add_rule(self, name, rule):
        esc_name = INVALID_RULE_CHARS_RE.sub("-", name)
        if esc_name not in self._rules or self._rules[esc_name] == rule:
            key = esc_name
        else:
            i = 0
            while f"{esc_name}{i}" in self._rules:
                i += 1
            key = f"{esc_name}{i}"
        self._rules[key] = rule
        return key

    def visit(self, schema, name):
        schema_type = schema.get("type")
        rule_name = name or "root"

        if "oneOf" in schema or "anyOf" in schema:
            rule = " | ".join(
                (
                    self.visit(alt_schema, f'{name}{"-" if name else ""}{i}')
                    for i, alt_schema in enumerate(
                        schema.get("oneOf") or schema["anyOf"]
                    )
                )
            )
            return self._add_rule(rule_name, rule)

        elif "const" in schema:
            return self._add_rule(
                rule_name, self._format_literal(schema["const"])
            )

        elif "enum" in schema:
            rule = " | ".join(
                (self._format_literal(v) for v in schema["enum"])
            )
            return self._add_rule(rule_name, rule)

        elif schema_type == "object" and "properties" in schema:
            # TODO: `required` keyword
            prop_order = self._prop_order
            prop_pairs = sorted(
                schema["properties"].items(),
                # sort by position in prop_order (if specified) then by key
                key=lambda kv: (prop_order.get(kv[0], len(prop_order)), kv[0]),
            )

            rule = '"{" space'
            for i, (prop_name, prop_schema) in enumerate(prop_pairs):
                prop_rule_name = self.visit(
                    prop_schema, f'{name}{"-" if name else ""}{prop_name}'
                )
                if i > 0:
                    rule += ' "," space'
                rule += rf' {self._format_literal(prop_name)} space ":" space {prop_rule_name}'
            rule += ' "}" space'

            return self._add_rule(rule_name, rule)

        elif schema_type == "array" and "items" in schema:
            # TODO `prefixItems` keyword
            item_rule_name = self.visit(
                schema["items"], f'{name}{"-" if name else ""}item'
            )
            rule = f'"[" space ({item_rule_name} ("," space {item_rule_name})*)? "]" space'
            return self._add_rule(rule_name, rule)

        else:
            assert (
                schema_type in PRIMITIVE_RULES
            ), f"Unrecognized schema: {schema}"
            return self._add_rule(
                "root" if rule_name == "root" else schema_type,
                PRIMITIVE_RULES[schema_type],
            )

    def format_grammar(self):
        return "\n".join(
            (f"{name} ::= {rule}" for name, rule in self._rules.items())
        )


def to_grammar(schema):
    converter = SchemaConverter({})
    converter.visit(schema, "")
    grammar_string = converter.format_grammar()
    return LlamaGrammar.from_string(grammar_string, verbose=False)



# -*- coding: utf-8 -*-
import os
from decouple import config

MODELS_CACHE_DIR = config(
    "MODELS_CACHE_DIR",
    default=os.path.expanduser("~/.cache/py-llm-core/models"),
)

if not os.path.exists(MODELS_CACHE_DIR):
    os.makedirs(MODELS_CACHE_DIR)

#: Store models in the same directory to simplify purges
os.environ["TIKTOKEN_CACHE_DIR"] = MODELS_CACHE_DIR


#: OpenAI
OPENAI_API_KEY = config("OPENAI_API_KEY", default=None)

USE_AZURE_OPENAI = config("USE_AZURE_OPENAI", cast=bool, default=False)

AZURE_OPENAI_ENDPOINT = config("AZURE_OPENAI_ENDPOINT", default=None)


#: Mistral
MISTRAL_API_KEY = config("MISTRAL_API_KEY", default=None)

USE_AZURE_AI_MISTRAL_LARGE = config(
    "USE_AZURE_AI_MISTRAL_LARGE", cast=bool, default=False
)

AZURE_AI_MISTRAL_LARGE_ENDPOINT = config(
    "AZURE_AI_MISTRAL_LARGE_ENDPOINT", default=None
)

AZURE_AI_MISTRAL_LARGE_KEY = config("AZURE_AI_MISTRAL_LARGE_KEY", default=None)


DEFAULT_TIMEOUT = config("DEFAULT_TIMEOUT", default=300.0)



# -*- coding: utf-8 -*-
from dataclasses import dataclass
from typing import List
import codecs
import json


@dataclass
class LLMBase:
    def sanitize_prompt(
        self,
        prompt,
        history=None,
        schema=None,
    ):
        schema_prompt = ""

        if schema:
            schema_prompt = json.dumps(schema)

        complete_prompt = [
            self.system_prompt,
            prompt,
            str(history) if history else "",
            schema_prompt,
        ]

        complete_prompt = "\n".join(complete_prompt)

        required_ctx_size = len(codecs.encode(complete_prompt, self.name))
        if required_ctx_size > self.ctx_size:
            raise OverflowError(
                f"Prompt too large {required_ctx_size} for this model {self.ctx_size}"
            )

        return self.ctx_size - required_ctx_size


@dataclass
class Usage:
    prompt_tokens: int
    completion_tokens: int
    total_tokens: int


@dataclass
class Message:
    role: str
    content: str
    function_call: dict = None
    tool_calls: dict = None
    name: str = None


@dataclass
class ChatCompletionChoice:
    index: int
    message: Message
    finish_reason: str

    @classmethod
    def from_iterable(cls, iterable):
        for item in iterable:
            message_attrs = item["message"]

            # To ensure compatibility with other models, we stuff
            # function calling in the message content.

            function_call = message_attrs.get("function_call")
            tool_calls = message_attrs.get("tool_calls")
            if function_call:
                message_attrs["content"] = message_attrs["function_call"][
                    "arguments"
                ]
            elif tool_calls:
                message_attrs["content"] = message_attrs["tool_calls"][0][
                    "function"
                ]["arguments"]

            message = Message(**item["message"])
            index = item["index"]
            finish_reason = item["finish_reason"]
            yield cls(
                index=index, message=message, finish_reason=finish_reason
            )


@dataclass
class ChatCompletion:
    id: str
    object: str
    created: int
    model: str
    choices: List[ChatCompletionChoice]
    usage: Usage
    system_fingerprint: str = None
    prompt_filter_results: dict = None

    @classmethod
    def parse(cls, attrs):
        attributes = {}
        attributes.update(attrs)

        attributes["choices"] = list(
            ChatCompletionChoice.from_iterable(attributes["choices"])
        )
        attributes["usage"] = Usage(**attributes["usage"])

        return cls(**attributes)



# -*- coding: utf-8 -*-
from typing import List
from dataclasses import dataclass


@dataclass
class QuestionCollection:
    system_prompt = "You are a helpful assistant."

    prompt = """Instructions:
    ```
    {instructions}
    ```

    The provided answer was:
    ```
    {answer}
    ```

    Write {n_questions} single-fact, close-ended questions to help verify
    if there are mistakes in the answer.
    """

    questions: List[str]


@dataclass
class Doubter:
    model: str
    assistant_cls: type
    results_cls: type = QuestionCollection

    def verify(self, instructions, answer, n_questions=5):
        with self.assistant_cls(
            self.results_cls, model=self.model
        ) as assistant:
            verification = assistant.process(
                instructions=instructions,
                answer=answer,
                n_questions=n_questions,
            )
            return verification


@dataclass
class AnswerConsistency:
    system_prompt = (
        "You are a meticulous assistant checking for inconsistencies"
    )

    prompt = """Context:
    ```
    {context}
    ```

    Question:
    ```
    {question}
    ```

    The provided Answer was:
    ```
    {answer}
    ```

    - Does the Answer is consistent ?
    - Does the Answer is inferred from the Context ?
    """

    is_consistent: bool
    is_inferred_from_context: bool


@dataclass
class ConsistencyVerifier:
    model: str
    assistant_cls: type
    results_cls: type = AnswerConsistency

    def verify(self, question, context, answer):
        with self.assistant_cls(
            self.results_cls, model=self.model
        ) as assistant:
            verification = assistant.process(
                question=question,
                context=context,
                answer=answer,
            )
            return verification



from .openai import OpenAIChatModel
from .llama_cpp_compatible import LLaMACPPModel, LLaVACPPModel
from .mistralai import MistralAILarge

__all__ = [
    "OpenAIChatModel",
    "LLaMACPPModel",
    "LLaVACPPModel",
    "MistralAILarge",
]



# -*- coding: utf-8 -*-
from dataclasses import dataclass


@dataclass
class Answer:
    system_prompt = "You are a helpful analyst."

    prompt = """Context:
    ```
    {context}
    ```

    Using *only* the Context, answer briefly to the following:
    ```
    {question}
    ```
    """

    content: str


@dataclass
class Analyst:
    model: str
    assistant_cls: type
    results_cls: type = Answer

    def batch_ask(self, questions, context):
        for question in questions:
            with self.assistant_cls(
                self.results_cls, model=self.model
            ) as assistant:
                answer = assistant.process(question=question, context=context)

            # Release before yielding
            yield answer

    def ask(self, question, context):
        with self.assistant_cls(
            self.results_cls, model=self.model
        ) as assistant:
            answer = assistant.process(question=question, context=context)
            return answer



# -*- coding: utf-8 -*-
import httpx
from openai import AzureOpenAI, OpenAI

from dataclasses import dataclass

from .base import (
    LLMBase,
    ChatCompletion,
)

from ..settings import (
    USE_AZURE_OPENAI,
    AZURE_OPENAI_ENDPOINT,
    DEFAULT_TIMEOUT,
)


def create_chat_completion(
    model, messages, temperature, max_tokens=1000, **kwargs
):
    default_timeout = httpx.Timeout(DEFAULT_TIMEOUT, write=10.0, connect=2.0)

    if USE_AZURE_OPENAI:
        # gets the API Key from environment variable AZURE_OPENAI_API_KEY
        client = AzureOpenAI(
            api_version="2024-02-01",
            azure_endpoint=AZURE_OPENAI_ENDPOINT,
            timeout=default_timeout,
        )

        # Here we are using the following convention:
        # Every model name is mapped to an Azure deployment where we remove the
        # dot sign.
        model_name = model.replace(".", "")
    else:
        client = OpenAI(timeout=default_timeout)
        model_name = model

    completion = client.chat.completions.create(
        model=model_name,
        messages=messages,
        temperature=temperature,
        max_tokens=max_tokens,
        **kwargs,
    )

    return completion


@dataclass
class OpenAIChatModel(LLMBase):
    name: str = "gpt-3.5-turbo"
    system_prompt: str = "You are a helpful assistant"

    def __enter__(self):
        # No special initialization required as we are using API
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # No special cleanup required as we are using API
        pass

    @property
    def ctx_size(self):
        if self.name in ("gpt-3.5-turbo", "gpt-3.5-turbo-0613"):
            return 4_000
        elif self.name in ("gpt-3.5-turbo-16k", "gpt-3.5-turbo-16k-0613"):
            return 16_000
        elif self.name in ("gpt-4", "gpt-4-0613"):
            return 8_000
        elif self.name == "gpt-4-32k":
            return 32_000
        elif self.name == "gpt-4-1106-preview":
            return 128_000
        else:
            raise KeyError("Unsupported model")

    def ask(
        self,
        prompt,
        history=None,
        schema=None,
        temperature=0,
    ):
        max_tokens = self.sanitize_prompt(
            prompt=prompt,
            history=history,
            schema=schema,
        )

        #: Reduce by 10 percent the maximum tokens to be generated to take into
        #: account inaccuracies of sanitize_prompt (especially the schema token
        #: consumption)

        max_tokens = int(0.9 * max_tokens)

        messages = [
            {
                "role": "system",
                "content": self.system_prompt,
            },
        ]
        if history:
            messages += history

        messages.append(
            {
                "role": "user",
                "content": prompt,
            },
        )

        kwargs = {}
        if schema:
            functions = {
                "name": "PublishAnswer",
                "description": "Publish the answer",
                "parameters": schema,
            }
            function_call = {"name": "PublishAnswer"}

            kwargs = {
                "functions": [functions],
                "function_call": function_call,
            }

        completion = create_chat_completion(
            model=self.name,
            messages=messages,
            temperature=temperature,
            max_tokens=max_tokens,
            **kwargs,
        )

        return ChatCompletion.parse(completion.dict())



# -*- coding: utf-8 -*-
import itertools
import codecs

from dataclasses import dataclass


@dataclass
class TokenSplitter:
    model: str = "gpt-3.5-turbo"
    chunk_size: int = 300
    chunk_overlap: int = 0

    def __post_init__(self):
        if self.chunk_overlap > self.chunk_size:
            raise IndexError(
                f"Overlap {self.chunk_overlap} > window size {self.chunk_size}"
            )

    def first_extract(self, text: str):
        return next(self.chunkify(text))

    def compute_token_count(self, text: str):
        return len(codecs.encode(text, self.model))

    def chunkify(self, text: str):
        tokens = codecs.encode(text, self.model)

        start = 0
        length = len(tokens)

        while True:
            stop = start + self.chunk_size
            chunk = list(itertools.islice(tokens, start, stop))
            yield codecs.decode(chunk, self.model)

            if stop >= length:
                break

            start += self.chunk_size - self.chunk_overlap



# -*- coding: utf-8 -*-
import os
import platform
import llama_cpp

from dataclasses import dataclass

from .base import (
    LLMBase,
    ChatCompletion,
)
from ..schema import to_grammar
from ..settings import MODELS_CACHE_DIR


@dataclass
class LLaMACPPModel(LLMBase):
    """
    Model wrapper using the https://github.com/ggerganov/llama.cpp library

    Not limited to LLaMA models. For the complete list of supported models,

    see https://github.com/ggerganov/llama.cpp#readme
    """

    name: str = "mistral"
    system_prompt: str = "You are a helpful assistant"
    ctx_size: int = 8000
    verbose: bool = False
    llama_cpp_kwargs: dict = None

    def __enter__(self):
        self.load_model()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.release_model()

    def release_model(self):
        del self.model

    def load_model(self):
        if self.llama_cpp_kwargs is None:
            self.llama_cpp_kwargs = {
                "n_ctx": self.ctx_size,
                "verbose": self.verbose,
            }

            if platform.system() == "Darwin" and platform.machine() == "arm64":
                # Offload everything onto the GPU on MacOS
                self.llama_cpp_kwargs["n_gpu_layers"] = 100
                self.llama_cpp_kwargs["n_threads"] = 1

        model_path = os.path.join(MODELS_CACHE_DIR, self.name)
        self.model = llama_cpp.Llama(model_path, **self.llama_cpp_kwargs)

    def ask(
        self,
        prompt,
        history=None,
        schema=None,
        temperature=0,
        **llama_cpp_kwargs,
    ):
        self.sanitize_prompt(prompt=prompt, history=history, schema=schema)

        messages = [
            {
                "role": "system",
                "content": self.system_prompt,
            },
        ]
        if history:
            messages += history

        messages.append(
            {
                "role": "user",
                "content": prompt,
            },
        )

        kwargs = {}
        if schema:
            grammar = to_grammar(schema)
            kwargs = {"grammar": grammar}

        # Allow to call `ask` and free up memory immediately
        model = getattr(self, "model", None)

        if model:
            completion = model.create_chat_completion(
                messages=messages,
                temperature=temperature,
                max_tokens=-1,
                **kwargs,
            )
        else:
            with self as wrapper:
                completion = wrapper.model.create_chat_completion(
                    messages=messages,
                    temperature=temperature,
                    **kwargs,
                )

        return ChatCompletion.parse(completion)


@dataclass
class LLaVACPPModel(LLaMACPPModel):
    def load_model(self):
        if self.llama_cpp_kwargs is None:
            self.llama_cpp_kwargs = {
                "n_ctx": self.ctx_size,
                "verbose": self.verbose,
            }

            if platform.system() == "Darwin" and platform.machine() == "arm64":
                # Offload everything onto the GPU on MacOS
                self.llama_cpp_kwargs["n_gpu_layers"] = 100
                self.llama_cpp_kwargs["n_threads"] = 1

        model_path = os.path.join(MODELS_CACHE_DIR, self.name)
        chat_format = "llava-1-5"
        chat_handler = llama_cpp.llama_chat_format.Llava15ChatHandler(
            clip_model_path=self.llama_cpp_kwargs["clip_model_path"]
        )

        self.model = llama_cpp.Llama(
            model_path,
            chat_format=chat_format,
            chat_handler=chat_handler,
            **self.llama_cpp_kwargs,
        )

    def ask(
        self,
        prompt,
        history=None,
        schema=None,
        temperature=0,
        **llama_cpp_kwargs,
    ):
        self.sanitize_prompt(prompt=prompt, history=history, schema=schema)

        messages = [
            {
                "role": "system",
                "content": self.system_prompt,
            },
        ]
        if history:
            messages += history

        messages.append(
            {
                "role": "user",
                "content": [{"type": "text", "text": prompt}],
            },
        )

        kwargs = {}
        if schema:
            grammar = to_grammar(schema)
            kwargs = {"grammar": grammar}

        # Allow to call `ask` and free up memory immediately
        model = getattr(self, "model", None)

        if model:
            completion = model.create_chat_completion(
                messages=messages,
                temperature=temperature,
                max_tokens=-1,
                **kwargs,
            )
        else:
            with self as wrapper:
                completion = wrapper.model.create_chat_completion(
                    messages=messages,
                    temperature=temperature,
                    **kwargs,
                )

        return ChatCompletion.parse(completion)



# coding: utf-8
from pathlib import Path
from setuptools import setup, find_packages


here = Path(__file__).parent
packages = find_packages("src")
main_package = packages[0]
long_description = (here / "README.md").read_text()
requirements = (here / "requirements.txt").read_text().splitlines()


setup(
    name="py-llm-core",
    version="2.8.4",
    license="MIT",
    description="PyLLMCore provides a light-weighted interface with LLMs",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="P.A. SCHEMBRI",
    author_email="pa.schembri@advanced-stack.com",
    url="https://github.com/paschembri/py-llm-core",
    packages=packages,
    package_dir={"": "src"},
    include_package_data=True,
    install_requires=requirements,
)



# -*- coding: utf-8 -*-
import codecs
import tiktoken
import llama_cpp
import os

from .llm import LLaMACPPModel

from .settings import MODELS_CACHE_DIR

codec_to_model_name = {}
for model_name in os.listdir(MODELS_CACHE_DIR):
    key = model_name.replace("-", "_").lower()
    codec_to_model_name[key] = model_name


class BaseTokenCodec:
    def encode(self, input, errors="strict"):
        output = self.token_encoder.encode(input)
        return output, len(output)

    def decode(self, input, errors="strict"):
        output = self.token_encoder.decode(input)
        return output, len(output)


class OpenAIGPTCodec(BaseTokenCodec):
    @classmethod
    def is_supported(cls, codec_name):
        return codec_name in (
            "gpt3.5",
            "gpt3.5_0613",
            "gpt_3.5",
            "gpt_3.5_0613",
            "gpt_3.5_turbo",
            "gpt_3.5_turbo_0613",
            "gpt_3.5_turbo_16k",
            "gpt_3.5_turbo_16k_0613",
            "gpt_4",
            "gpt_4_0613",
            "gpt_4_1106_preview",
            "gpt_4_32k",
            "mistral_large_latest",  #: dirty hack to quickly support mistral
            "azureai",  #: dirty hack to quickly support mistral
        )

    def __init__(self, codec_name):
        # Map everything to gpt-3.5-turbo
        self.token_encoder = tiktoken.encoding_for_model("gpt-3.5-turbo")


class LLaMACodec:
    @classmethod
    def is_supported(cls, codec_name):
        return codec_name in codec_to_model_name

    def __init__(self, codec_name):
        self.model_name = codec_to_model_name[codec_name]

    def encode(self, input, errors="strict"):
        with LLaMACPPModel(name=self.model_name) as wrapper:
            tokenizer = llama_cpp.LlamaTokenizer(wrapper.model)
            output = tokenizer.encode(input)

        return output, len(output)

    def decode(self, input, errors="strict"):
        with LLaMACPPModel(name=self.model_name) as wrapper:
            tokenizer = llama_cpp.LlamaTokenizer(wrapper.model)
            output = tokenizer.decode(input)

        return output, len(output)


def lookup(codec_name):
    if OpenAIGPTCodec.is_supported(codec_name):
        codec = OpenAIGPTCodec(codec_name)

    elif LLaMACodec.is_supported(codec_name):
        codec = LLaMACodec(codec_name)

    else:
        return None

    return codecs.CodecInfo(
        name=codec_name,
        encode=codec.encode,
        decode=codec.decode,
    )



# -*- coding: utf-8 -*-
import dirtyjson

from .llm import OpenAIChatModel, LLaMACPPModel, MistralAILarge
from .schema import to_json_schema, from_dict


class BaseParser:
    def __init__(self, target_cls, *args, **kwargs):
        self.target_cls = target_cls
        self.target_json_schema = to_json_schema(self.target_cls)

    def deserialize(self, json_str):
        attributes = dirtyjson.loads(json_str)
        return from_dict(self.target_cls, attributes)

    def parse(self, text):
        prompt = "Extract and parse information from provided content"
        history = [{"role": "user", "content": text}]
        completion = self.model_wrapper.ask(
            prompt, history, schema=self.target_json_schema
        )
        instance = self.deserialize(completion.choices[0].message.content)
        return instance

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class OpenAIParser(BaseParser):
    def __init__(
        self,
        target_cls,
        model="gpt-3.5-turbo",
        completion_kwargs=None,
        *args,
        **kwargs
    ):
        super().__init__(target_cls, *args, **kwargs)
        self.completion_kwargs = (
            {} if completion_kwargs is None else completion_kwargs
        )

        self.model_wrapper = OpenAIChatModel(
            name=model,
            system_prompt=(
                "Act as a powerful AI able to extract, parse and process "
                "information from unstructured content."
            ),
        )
        self.ctx_size = self.model_wrapper.ctx_size
        self.model_name = self.model_wrapper.name


class LLaMACPPParser(BaseParser):
    """
    Parser using the https://github.com/ggerganov/llama.cpp library

    Not limited to LLaMA models. For the complete list of supported models,

    see https://github.com/ggerganov/llama.cpp#readme
    """

    def __init__(
        self,
        target_cls,
        model="mistral",
        completion_kwargs=None,
        llama_cpp_kwargs=None,
        *args,
        **kwargs
    ):
        super().__init__(target_cls, *args, **kwargs)
        self.completion_kwargs = (
            {} if completion_kwargs is None else completion_kwargs
        )

        self.model_wrapper = LLaMACPPModel(
            name=model, llama_cpp_kwargs=llama_cpp_kwargs
        )
        self.ctx_size = self.model_wrapper.ctx_size
        self.model_name = self.model_wrapper.name

    def __enter__(self):
        self.model_wrapper.load_model()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.model_wrapper.release_model()


class MistralAILargeParser(BaseParser):
    def __init__(
        self,
        target_cls,
        model="mistral-large-latest",
        completion_kwargs=None,
        *args,
        **kwargs
    ):
        super().__init__(target_cls, *args, **kwargs)
        self.completion_kwargs = (
            {} if completion_kwargs is None else completion_kwargs
        )

        self.model_wrapper = MistralAILarge(
            name=model,
            system_prompt=(
                "Act as a powerful AI able to extract, parse and process "
                "information from unstructured content."
            ),
        )
        self.ctx_size = self.model_wrapper.ctx_size
        self.model_name = self.model_wrapper.name



# -*- coding: utf-8 -*-
from typing import List
from dataclasses import dataclass
from ..splitters import TokenSplitter


@dataclass
class Summary:
    content: str


@dataclass
class SimpleSummary(Summary):
    system_prompt = """
    You are an expert at extracting key information and summarizing.
    """

    prompt = """Content:
    ```
    {content}
    ```
    --
    Summarize the content in {word_count}.
    """


@dataclass
class DenseSummary(Summary):
    missing_entities: List[str]


@dataclass
class DenserSummaryCollection:
    system_prompt = """
    You are an expert in writing rich and dense summaries in broad domains.
    """

    prompt = """Content:
    ```
    {content}
    ```
    --

    You will generate increasingly concise, entity-dense summaries of the above
    Article.

    Repeat the following 2 steps 5 times.

    - Step 1: Identify 1-3 informative Entities from the Article
    which are missing from the previously generated summary and are the most
    relevant.

    - Step 2: Write a new, denser summary of identical length which covers
    every entity and detail from the previous summary plus the missing entities

    A Missing Entity is:

    - Relevant: to the main story
    - Specific: descriptive yet concise (5 words or fewer)
    - Novel: not in the previous summary
    - Faithful: present in the Article
    - Anywhere: located anywhere in the Article

    Guidelines:
    - The first summary should be long (4-5 sentences, approx. {word_count} words) yet
    highly non-specific, containing little information beyond the entities
    marked as missing.

    - Use overly verbose language and fillers (e.g. "this article discusses") to
    reach approx. {word_count} words.

    - Make every word count: re-write the previous summary to improve flow and
    make space for additional entities.

    - Make space with fusion, compression, and removal of uninformative phrases
    like "the article discusses"

    - The summaries should become highly dense and concise yet self-contained,
    e.g., easily understood without the Article.

    - Missing entities can appear anywhere in the new summary.

    - Never drop entities from the previous summary. If space cannot be made,
    add fewer new entities.

    > Remember to use the exact same number of words for each summary.
    Answer in JSON.

    > The JSON in `summaries` should be a list (length 5) of
    dictionaries whose keys are "missing_entities" and "content".

    """

    summaries: List[DenseSummary]


@dataclass
class Summarizer:
    model: str
    assistant_cls: type
    results_cls: type = SimpleSummary

    def fast_summarize(self, content, word_count=80):
        with self.assistant_cls(
            self.results_cls, model=self.model
        ) as assistant:
            splitter = TokenSplitter(
                model=assistant.model_name,
                chunk_size=int(assistant.ctx_size * 0.6),
                chunk_overlap=int(assistant.ctx_size * 0.05),
            )
            chunk = next(splitter.chunkify(content))
            summary = assistant.process(content=chunk, word_count=word_count)
            return summary

    def summarize(self, content, word_count=80):
        with self.assistant_cls(
            self.results_cls, model=self.model
        ) as assistant:
            splitter = TokenSplitter(
                model=assistant.model_name,
                chunk_size=int(assistant.ctx_size * 0.6),
                chunk_overlap=int(assistant.ctx_size * 0.05),
            )
            for chunk in splitter.chunkify(content):
                summary = assistant.process(
                    content=chunk, word_count=word_count
                )
                yield summary
